<?php
/**
 * Plugin Weiqi: Displays Weiqi (also named Go game or Baduk) diagrams.
 * http://wiki.splitbrain.org/plugin:weiqi
 * http://en.wikipedia.org/wiki/Go_(board_game)
 *
 * @license    GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author     Christophe Gragnic <christophegragnic@yahoo.fr>
 */
 
// must be run within Dokuwiki
if(!defined('DOKU_INC')) die();
 
if(!defined('DOKU_PLUGIN')) define('DOKU_PLUGIN',DOKU_INC.'lib/plugins/');
require_once(DOKU_PLUGIN.'syntax.php');
 
/**
 * All DokuWiki plugins to extend the parser/rendering mechanism
 * need to inherit from this class
 */
class syntax_plugin_weiqi extends DokuWiki_Syntax_Plugin {

    // these class vars are like conf vars (will be in the admin page one day)
    // they are initialized in the plugin constructor syntax_plugin_weiqi()
    
    // syntax starts at <xxx>, ends at </xxx> where xxx is $syntax_tag
    var $syntax_tag;
    // how to know a line is for the attributes
    var $attributes_key;
    // how to know a line is for the caption
    var $caption_key;
    // web img path (with trailing slash)
    var $img_path;
    // letter sequence (no 'i' on a goban)
    var $letter_sequence;
    // when displaying plain text instead of images,
    // if font-size:(image_size)px; it's too big, hence we have to reduce it
    var $plain_text_coeff;
    
    /**
     * Plugin Constructor.
     */
    function syntax_plugin_weiqi() {
      // enable direct access to language strings
      $this->setupLocale();
      // enable direct access to configuration
      $this->loadConfig();
      
      $this->syntax_tag = 'weiqi';
      $this->attributes_key = 'attributes:';
      $this->caption_key = 'caption:';
      $this->img_path = DOKU_BASE.'lib/plugins/weiqi/img/';
      $this->letter_sequence = 'abcdefghjklmnopqrstuvwxyz';
      $this->plain_text_coeff = 0.7;
    }

    function getInfo() {
      return array(
        'author' => 'Grahack',
        'email'  => 'christophegragnic@yahoo.fr',
        'date'   => '2008-04-26',
        'name'   => 'Weiqi Plugin',
        'desc'   => 'Displays Weiqi (also named Go game or Baduk) diagrams',
        'url'    => 'http://wiki.splitbrain.org/plugin:weiqi',
      );
    }
    
    function getType() { return 'protected'; }
    function getAllowedTypes() { return array('formatting', 'disabled'); }
    function getSort() { return 999; }
    function connectTo($mode) {
        $pattern = '<'.$this->syntax_tag.'.*?>(?=.*?</'.$this->syntax_tag.'>)';
        $this->Lexer->addEntryPattern($pattern,$mode,'plugin_weiqi');
    }
    function postConnect() {
        $this->Lexer->addExitPattern('</'.$this->syntax_tag.'>','plugin_weiqi');
    }
    function handle($match, $state, $pos, &$handler) {
        if ($state == DOKU_LEXER_UNMATCHED) {
            $lines = explode("\n", $match);
            // init of attributes, caption and bi-dimensional array of chars
            $attributes_str = '';
            $caption = '';
            $board = array();
            
            // first pass for some checks, caption and attributes
            foreach ($lines as $line) {
                if ($line!='') {
                    // check if this is the attributes line
                    // maybe preg_match is better
                    if (substr($line, 0, strlen($this->attributes_key)) == $this->attributes_key) {
                        $attributes_str.= ' '.substr($line, strlen($this->attributes_key));
                    // check if this is the caption line
                    } elseif (substr($line, 0, strlen($this->caption_key)) == $this->caption_key) {
                        if ($caption!='')
                            return array($state, array(), $this->lang['weiqi one caption']);
                        $caption = trim(substr($line, strlen($this->caption_key)));
                    } else {
                        // check length of our lines
                        // first store the one of the first line
                        if (empty($line_length)) $line_length = strlen($line);
                        if (strlen($line) != $line_length)
                            return array($state, array(), $this->lang['weiqi rectangular']);
                    }
                }
            }
            $attributes = $this->parse_goban_attributes($attributes_str);
            if (!is_array($attributes)) {
                $message = $this->lang['weiqi bad attributes'].' ('.trim($attributes_str).')<br />';
                $message.= $attributes;
                return array($state, array(), $message);
            }
            
            // second pass for parsing the board code
            foreach ($lines as $line) {
                if ($line!='') {
                    // check if this is the attributes or caption line
                    $att = substr($line, 0, strlen($this->attributes_key)) == $this->attributes_key;
                    $cap = substr($line, 0, strlen($this->caption_key)) == $this->caption_key;
                    if ( !$att AND !$cap) {
                        $tmp_array_line = array();
                        // map the string to an array
                        // one by one for standard code,
                        // two by two for advanced code
                        if (!$attributes['advanced']) {
                            for($i=0;$i<strlen($line);$i++)
                                $tmp_array_line[] = substr($line, $i, 1);
                        } else {
                            for($i=0;$i<strlen($line)/2;$i++)
                                $tmp_array_line[] = substr($line, 2*$i, 2);
                        }
                        // append this line to the main bi-dim array of chars
                        $board[] = $tmp_array_line;
                    }
                }
            }
            $width = $attributes['advanced']?($line_length/2):$line_length;
            $height = count($board);
            
            return array($state, array($attributes, $caption, $board, $width, $height, $match));
        } else return array($state);
    }
    function render($mode, &$renderer, $data) {
        if ($mode == 'xhtml') {
            list($state, $goban_data, $message) = $data;
            switch ($state) {
                case DOKU_LEXER_ENTER :
                    $renderer->doc .= '<div class="weiqi">';
                    return true;
                case DOKU_LEXER_UNMATCHED :
                    if ($message!='') {
                        $renderer->doc.= '<div class="error">';
                        $renderer->doc.= $this->lang['weiqi error'].': '.$message;
                        $renderer->doc.= '</div>';
                        // no need to display the goban if something is wrong
                        // so we stop here
                        return true;
                    }
                    list($attributes, $caption, $board, $width, $height, $source) = $goban_data;
                    // default settings
                    $demo = false;
                    $goban = 'images/wood1';
                    $w1 = 25;
                    $w2 = 10;
                    $coords = false;
                    $reverse_numbers = false;
                    $reverse_letters = false;
                    $start_number = 1;
                    $start_letter = 1;
                    $advanced = false;
                    // import user-defined attributes
                    // quite dangerous, but controlled by $allowed_keys
                    // in the parse_goban_attributes() function
                    foreach ($attributes as $key => $val) $$key = $val;
                    
                    $wout = ''; // weiqi output, appended to $renderer->doc at the end
                    
                    // should we output the source too (demo purpose)
                    if ($demo) {
                        $wout.= '<pre class="w-source">';
                        $wout.= '&lt;'.$this->syntax_tag.'&gt;';
                        $wout.= $source;
                        $wout.= '&lt;/'.$this->syntax_tag.'&gt;';
                        $wout.= '</pre>';
                    }
                    
                    // hardcoded style (I mean not CSSed) to be able to change the goban color
                    $wout.= '<table style="background-image: url(\''.$this->img_path.$goban.'.gif\');" border="0" cellspacing="0" cellpadding="0">';
                    if ($caption!='') $wout .= '<caption>'.$caption.'</caption>';
                    // top coordinates
                    if ($coords) {
                        $wout .= '<tr>';
                        $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w1).'</td>';// blank for left coordinates
                        $wout .= '<td>'.$this->img('images/blank',' ',$w2,$w1).'</td>';// blank for left edge
                        for($x=1;$x<=$width;$x++) {
                            if ($reverse_letters) $row_letter = $this->letter($width - ($x - ($start_letter - 1)) + 1);
                            else $row_letter = $this->letter($x + ($start_letter - 1));
                            $wout .= '<td>'.$this->img($w1.'/c'.$row_letter, $row_letter).'</td>';// top coordinates
                        }
                        $wout .= '<td>'.$this->img('images/blank',' ',$w2,$w1).'</td>';// blank for right edge
                        $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w1).'</td>';// blank for right coordinates
                    $wout .= '</tr>';
                    }
                    // top edge
                    $wout .= '<tr>';
                    if ($coords) $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w2).'</td>';// blank for left coordinates
                    $wout .= '<td>'.$this->img($goban.'_ul', ' ').'</td>';// top left goban corner
                    $wout .= '<td colspan="'.$width.'">'.$this->img($goban.'_u', ' ', $w1*$width, $w2).'</td>';// top corner
                    $wout .= '<td>'.$this->img($goban.'_ur', ' ').'</td>';// top right goban corner
                    if ($coords) $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w2).'</td>';// blank for right coordinates
                    $wout .= '</tr>';
                    // body
                    $y=0;
                    foreach ($board as $line) {
                        $y++;
                        $x=0;
                        $wout .= '<tr>';
                        if ($reverse_numbers) $line_number = $y + ($start_number - 1);
                        else $line_number = $height - $y + 1 + ($start_number - 1);
                        
                        if ($coords) $wout .= '<td>'.$this->img($w1.'/c'.$line_number, $line_number).'</td>';// left coordinates
                        if ($y==1) $wout .= '<td rowspan="'.$height.'">'.$this->img($goban.'_l', ' ', $w2, $w1*$height).'</td>';// left edge
                        // board
                        foreach ($line as $char) {
                            $x++;
                            $wout .= '<td>'.$this->img($w1.'/'.$this->weiqi_code($char, $x, $y, $width, $height), $char, $w1).'</td>';
                        }
                        if ($y==1) $wout .= '<td rowspan="'.$height.'">'.$this->img($goban.'_r', ' ', $w2, $w1*$height).'</td>';// right edge
                        if ($coords) $wout .= '<td>'.$this->img($w1.'/c'.$line_number, $line_number).'</td>';// left coordinates
                        $wout .= '</tr>';
                    }
                    // bottom edge
                    $wout .= '<tr>';
                    if ($coords) $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w2).'</td>';// blank for left coordinates
                    $wout .= '<td>'.$this->img($goban.'_dl', ' ').'</td>';// bottom left goban corner
                    $wout .= '<td colspan="'.$width.'">'.$this->img($goban.'_d', ' ', $w1*$width, $w2).'</td>';
                    $wout .= '<td>'.$this->img($goban.'_dr', ' ').'</td>';// bottom right goban corner
                    if ($coords) $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w2).'</td>';// blank for right coordinates
                    $wout .= '</tr>';
                    // bottom coordinates
                    if ($coords) {
                        $wout .= '<tr>';
                        $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w1).'</td>';// blank for left coordinates
                        $wout .= '<td>'.$this->img('images/blank',' ',$w2,$w1).'</td>';// blank for left edge
                        for($x=1;$x<=$width;$x++) {
                            if ($reverse_letters) $row_letter = $this->letter($width - ($x - ($start_letter - 1)) + 1);
                            else $row_letter = $this->letter($x + ($start_letter - 1));
                            $wout .= '<td>'.$this->img($w1.'/c'.$row_letter, $row_letter).'</td>';// bottom coordinates
                        }
                        $wout .= '<td>'.$this->img('images/blank',' ',$w2,$w1).'</td>';// blank for right edge
                        $wout .= '<td>'.$this->img('images/blank',' ',$w1,$w1).'</td>';// blank for right coordinates
                        $wout .= '</tr>';
                    }
                    $wout.= '</table>';
                    $renderer->doc .= $wout;
                    return true;
                case DOKU_LEXER_EXIT :
                    $renderer->doc .= "</div>";
                    return true;
            }
        }
        return false;
    }
    function weiqi_code($str, $x, $y, $width, $height) {
        $str = trim($str);
        if (strlen($str) == 1) {
            // normal code
            switch ($str) {
                // goban
                case '.': return 'e';
                case ',': return 'h';
                case '+':
                    if ( $x==$width  && $y==1        ) return 'ur';
                    if ( $x==1       && $y==1        ) return 'ul';
                    if ( $x==1       && $y==$height  ) return 'dl';
                    if ( $x==$width  && $y==$height  ) return 'dr';
                case '-':
                    if ( $y==1        ) return 'u';
                    if ( $y==$height  ) return 'd';
                case '|':
                    if ( $x==1       ) return 'el';
                    if ( $x==$width  ) return 'er';
                
                // stones
                case 'x': return 'b';
                case 'X': return 'bm';
                case 'o': return 'w';
                case 'O': return 'wm';
                
                // for letters or numbers, we append an '#'
                // for them to appear as plain text
                
                // letters and numbers
                case '1': return '1#';
                case '2': return '2#';
                case '3': return '3#';
                case '4': return '4#';
                case '5': return '5#';
                case '6': return '6#';
                case '7': return '7#';
                case '8': return '8#';
                case '9': return '9#';
                    
                case 'a': return 'a#';
                case 'b': return 'b#';
                case 'c': return 'c#';
                case 'd': return 'd#';
                case 'e': return 'e#';
                case 'f': return 'f#';
                case 'g': return 'g#';
                case 'h': return 'h#';
                case 'i': return 'i#';
                case 'j': return 'j#';
                case 'k': return 'k#';
                case 'l': return 'l#';
                case 'm': return 'm#';
                case 'n': return 'n#';
                case 'p': return 'p#';
                case 'q': return 'q#';
                case 'r': return 'r#';
                case 's': return 's#';
                case 't': return 't#';
                case 'u': return 'u#';
                case 'v': return 'v#';
                case 'w': return 'w#';
                case 'y': return 'y#';
                case 'z': return 'z#';
                
                case 'A': return 'A#';
                case 'B': return 'B#';
                case 'C': return 'C#';
                case 'D': return 'D#';
                case 'E': return 'E#';
                case 'F': return 'F#';
                case 'G': return 'G#';
                case 'H': return 'H#';
                case 'I': return 'I#';
                case 'J': return 'J#';
                case 'K': return 'K#';
                case 'L': return 'L#';
                case 'M': return 'M#';
                case 'N': return 'N#';
                case 'P': return 'P#';
                case 'Q': return 'Q#';
                case 'R': return 'R#';
                case 'S': return 'S#';
                case 'T': return 'T#';
                case 'U': return 'U#';
                case 'V': return 'V#';
                case 'W': return 'W#';
                case 'Y': return 'Y#';
                case 'Z': return 'Z#';
            }
        } else {
            // advanced code
            // for letters or numbers, we append an '#'
            // for them to appear as plain text
            
            // the rest of the numbers
            if (is_numeric($str)) return $str.'#';
            else {
            // the rest of the available symbols
                $ch1 = substr($str, 0, 1);
                $ch2 = substr($str, 1, 1);
                // first, the rest of the letters
                if ($ch2 == 'l') return $ch1.'#';
                // now the goban marks and stones
                // ok it's a simple map, but DGS files for red squares have a d
                $ch2 = ($ch2=='r')?'d':$ch2;
                
                // some checks now
                // if nothing found, let's return nothing
                // so the img() function can report a mistake
                $available_second_chars = 'cstbwxdg';
                if (strpos($available_second_chars, $ch2) === false) return '';
                // no black stone with a black square, same for white
                if ($ch1.$ch2=='bb' OR $ch1.$ch2=='ww') return '';
                
                switch ($ch1) {
                    // goban and marks
                    case '.': return 'e'.$ch2;
                    case ',': return 'h'.$ch2;
                    case '+':
                        if ( $x==$width  && $y==1        ) return 'ur'.$ch2;
                        if ( $x==1       && $y==1        ) return 'ul'.$ch2;
                        if ( $x==1       && $y==$height  ) return 'dl'.$ch2;
                        if ( $x==$width  && $y==$height  ) return 'dr'.$ch2;
                    case '-':
                        if ( $y==1        ) return 'u'.$ch2;
                        if ( $y==$height  ) return 'd'.$ch2;
                    case '|':
                        if ( $x==1       ) return 'el'.$ch2;
                        if ( $x==$width  ) return 'er'.$ch2;
                    
                    // stones
                    case 'x': ;
                    case 'X': return 'b'.$ch2;
                    case 'o': ;
                    case 'O': return 'w'.$ch2;
                }
            }
        }
        // if nothing found, let's return nothing,
        // so the img() function can report a mistake
        return '';
    }
    function letter($n) {
        $n--;
        if ($n >= 0 AND $n < strlen($this->letter_sequence))
            return substr($this->letter_sequence, $n, 1);
        // this can help debugging the wiki code
        else return $n;
    }
    function img($src, $alt=false, $w=false, $h=false) {
        $text_px = floor($this->plain_text_coeff*$w);
        // '/' at the end means we have to report a mistake
        if (preg_match('@/$@', $src))
            return '<span class="w-report" style="font-size:'.$text_px.'px">@</span>';
        // '#' at the end means we have a number or a letter
        // numbers and letters are displayed with simple text
        if (preg_match('@/(.+)#$@', $src, $match))
            return '<span style="font-size:'.$text_px.'px">'.$match[1].'</span>';
        
        // the rest is done with the usual img html tag
        $html = '';
        $html.= '<img';
        if($w) $html.= ' width="'.$w.'"';
        if($h) $html.= ' height="'.$h.'"';
        $html.= ' src="'.$this->img_path.$src.'.gif"';
        if($alt) $html.= ' alt="'.$alt.'"';
        $html.= ' />';
        return $html;
    }
    function parse_goban_attributes($str) {
        $str = trim($str);
        if ($str=='') return array();
        
        $allowed_keys = array(
            'demo', 'goban', 'coords', 'grid_size',
            'reverse_numbers', 'reverse_letters',
            'start_number', 'start_letter',
            'advanced');
        
        $ret = array();
        $attributes = explode(' ', $str);
        $errors = '';
        foreach ($attributes as $attribute) {
            list($key,$value) = preg_split('/=/',$attribute,2);
            if (!in_array($key, $allowed_keys))
                $errors.= ', "'.$key.'" '.$this->lang['weiqi not allowed'];
            
            if ($key == 'demo') $ret[$key] = true;
            
            if ($key == 'goban') {
                if (!is_file(realpath('lib/plugins/weiqi/img/images/wood'.$value.'.gif'))) {
                    $errors.= ', "'.$value.'" '.$this->lang['weiqi not allowed for'].' '.$key;
                    $errors.= ' ('.$this->lang['weiqi not installed'].')';
                }
                $ret['goban'] = 'images/wood'.$value;
            }
            
            if ($key == 'coords') $ret[$key] = true;
            
            if ($key == 'grid_size') {
                if (!is_dir(realpath('lib/plugins/weiqi/img/'.$value))) {
                    $errors.= ', "'.$value.'" '.$this->lang['weiqi not allowed for'].' '.$key;
                    $errors.= ' ('.$this->lang['weiqi not installed'].')';
                }
                $ret['w1'] = $value;
            }
            
            if ($key == 'reverse_numbers') $ret[$key] = true;
            if ($key == 'reverse_letters') $ret[$key] = true;
            
            if ($key == 'start_number') {
                if ($value<1 OR $value>25) {
                    $errors.= ', "'.$value.'" '.$this->lang['weiqi not allowed for'].' '.$key;
                }
                $ret['start_number'] = $value;
            }
            if ($key == 'start_letter') {
                $strpos = strpos($this->letter_sequence, strtolower($value));
                if ($strpos===false) {
                    $errors.= ', "'.$value.'" '.$this->lang['weiqi not allowed for'].' '.$key;
                }
                $ret['start_letter'] = $strpos + 1;
            }
            
            if ($key == 'advanced') $ret[$key] = true;
        }
        if ($errors!='') return trim(trim($errors, ','));
        return $ret;
    }
}
